-- Copyright 2011-12 Paul Kulchenko, ZeroBrane LLC

local pathcache
local win = ide.osname == "Windows"
local mac = ide.osname == "Macintosh"

local function isValidPid(bid, cmd)
  if not bid or bid == -1 or bid == 0 then
    ide:Print(("Program unable to run as '%s'."):format(cmd))
    return
  end
  return bid
end

local function waitToComplete(bid)
  while wx.wxProcess.Exists(bid) do
    wx.wxSafeYield()
    wx.wxWakeUpIdle()
    wx.wxMilliSleep(100)
  end
end

return {
  name = "GiderosLink",
  description = "Gideros mobile platform",
  api = {"baselib", "gideros"},
  frun = function(self,wfilename,rundebug)
    local gideros = ide.config.path.gideros or pathcache -- check if the path is configured
    if not gideros then
      local sep = win and ';' or ':'
      local default =
           win and (GenerateProgramFilesPath('Gideros', sep)..sep)
        or mac and ('/Applications/Gideros Studio/Gideros Player.app/Contents/MacOS'..sep)
        or ''
      local path = default
                 ..(os.getenv('PATH') or '')..sep
                 ..(os.getenv('HOME') and os.getenv('HOME') .. '/bin' or '')
      local paths = {}
      for p in path:gmatch("[^"..sep.."]+") do
        gideros = gideros or GetFullPathIfExists(p, win and 'GiderosPlayer.exe' or 'Gideros Player')
        table.insert(paths, p)
      end
      if not gideros then
        ide:Print("Can't find gideros executable in any of the folders in PATH: "
          ..table.concat(paths, ", "))
        return
      end
      pathcache = gideros
    end
    if gideros and not wx.wxFileName(gideros):FileExists() then
      ide:Print("Can't find the specified gideros executable '"..gideros.."'.")
      return
    end

    local giderostools = wx.wxFileName.DirName(wx.wxFileName(gideros)
      :GetPath(wx.wxPATH_GET_VOLUME)
      ..(win and '/Tools' or '/../../../Gideros Studio.app/Contents/Tools'))
    giderostools:Normalize()
    local giderospath = giderostools:GetPath(wx.wxPATH_GET_VOLUME)
    local gdrbridge = GetFullPathIfExists(giderospath, win and 'gdrbridge.exe' or 'gdrbridge')
    if not gdrbridge then
      ide:Print("Can't find gideros bridge executable in '"..giderospath.."'.")
      return
    end

    -- find *.gproj file in the project directory
    local file
    for _, proj in ipairs(FileSysGetRecursive(self:fworkdir(wfilename), false, "*.gproj")) do
      if file then
        ide:Print("Found multiple .gproj files in the project directory; ignored '"..proj.."'.")
      end
      file = file or proj
    end
    if not file then
      ide:Print("Can't find gideros project file in the project directory.")
      return
    end

    if rundebug then
      ide:GetDebugger():SetOptions(
        {redirect = "c", runstart = ide.config.debugger.runonstart ~= false})
    end

    local pid
	
	local function runBridge(cmd,noret)
		cmd = ('"%s" %s'):format(gdrbridge, cmd)
		ide:Print("Starting '"..cmd.."'.")
		local proc = wx.wxProcess()
		proc:Redirect()
		proc:Connect(wx.wxEVT_END_PROCESS, function(event) proc = nil end)
		local bid = wx.wxExecute(cmd, wx.wxEXEC_ASYNC + wx.wxEXEC_MAKE_GROUP_LEADER, proc)
		local res
		if isValidPid(bid, cmd) then
			if noret then return true end
			local streamin = proc:GetInputStream()
			for _ = 1, 30 do
			  if streamin:CanRead() then
				res=streamin:Read(4096)
				if res and #res then
					local lz=res:find(string.char(0))
					if lz>0 then
						res=res:sub(1,lz-1)
					end
				end
				--ide:Print("RD ",res)
				break 
			  end
			  wx.wxSafeYield()
			  wx.wxWakeUpIdle()
			  wx.wxMilliSleep(250)
			end

			if res == nil and proc then
			  wx.wxProcess.Kill(bid, wx.wxSIGKILL, wx.wxKILL_CHILDREN)
			end
		end

		--ide:Print("Got ",res)
		return res
	end

    local remote = ide.config.gideros and ide.config.gideros.remote
    if remote then
      local cmd = ('"%s" %s "%s"'):format(gdrbridge, 'setip', remote)
      ide:Print(("Configuring remote player at %s."):format(remote))
      local bid = wx.wxExecute(cmd, wx.wxEXEC_ASYNC)
      if not isValidPid(bid, cmd) then return end
      waitToComplete(bid) -- wait for a bit to give Gideros chance to connect
    else
      ide:Print("Starting the player")
      local cmd = ('"%s"'):format(gideros)
      -- CommandLineRun(cmd,wdir,tooutput,nohide,stringcallback,uid,endcallback)
      pid = CommandLineRun(cmd,self:fworkdir(wfilename),not mac,true,nil,nil)
      if not pid then return end
    end

	require "json"
	local gapi=require "gideros_protocol"
	
    do
		ide:Print("Waiting for the bridge to connect at '"..gdrbridge.."'.")
		local cmd = ('"%s" %s'):format(gdrbridge, 'isconnected')
		local attempts, connected = 15
		for _ = 1, attempts do
			local res=runBridge('isconnected')
			connected = tonumber(res) == 1

			if connected then break end
			if connected == nil and proc then
			  if not remote then wx.wxProcess.Kill(pid, wx.wxSIGKILL, wx.wxKILL_CHILDREN) end
			  ide:Print("Couldn't connect to the player. Try again or check starting the player and the bridge manually.")
			  return
			end
		end
		if not connected then
			if not remote then wx.wxProcess.Kill(pid, wx.wxSIGKILL, wx.wxKILL_CHILDREN) end
			ide:Print("Couldn't connect after "..attempts.." attempts. Try again or check starting the player manually.")
			return
		end

		local loadres=runBridge(('%s "%s"'):format('load', file))
		ide:Print(("Starting project file '%s' => '%s'."):format(file,loadres))
		if not loadres then return end
		
		loadres=nil
		for _ = 1, 15 do
			local res=runBridge('loadstatus')
			if not res then return end
			ide:Print(("Waiting load: '%s'."):format(res))
			if res:sub(1,7)=="loaded:" then
				loadres=res:sub(8)
				break
			elseif res=="noconnect" or res=="" then
				ide:Print(("No connection to player."))
				return
			end
		end
		runBridge('stopdeamon',true)
		wx.wxMilliSleep(500)

		local files={}
		for p in loadres:gmatch("[^|]+") do 
			files[#files+1]=p 
		end
	  
		local playerIp=remote or "127.0.0.1"
		local client={}
		local gsock=gapi.connect(playerIp,15000,client)
		function client:print(s)
			s = s:gsub("%f[\r\n]%s*%*.-[\r\n]",""):gsub("^%s*%*.-[\r\n]+","")
			ide:GetOutput():Write(s)
		end
		function client:ack(ack)
			ide:Print("CommandAck:"..tostring(ack))
		end
		function client:connected()
			ide:Print("Player connected")
			gsock:play(files)
		end
		function client:disconnected()
			ide:Print("Player disconnected")
			gsock.timer:Stop()
		end
	  
		gsock.timer = ide:AddTimer(wx.wxGetApp(), 
		function()
			if pid and not wx.wxProcess.Exists(pid) then
				gsock:disconnect()
			end
			local _,err=pcall(gsock.poll,gsock) 
			if err then ide:Print("Polling:"..(tostring(err) or "")) end 			
		end)
		gsock.timer:Start(100)
    end
    return pid
  end,
  hasdebugger = true,
  scratchextloop = true,
  skipcompile = true, -- skip compilation as Gideros 2016.08+ includes bitops and macros
}
